# Import pingouin objects
from .utils import *
from .bayesian import *
from .effsize import *
from .multicomp import *
from .parametric import *
from .nonparametric import *
from .correlation import *
from .circular import *
from .pairwise import *
from .power import *

__version__ = "0.1.5"
